/* kas_convfs.c - A program to handle "KAS" MacSCII <-> ISO 8859-1
                  charset conversions for filenames for Xinets
		  K/A-Share server products.
		  
   Copyright (c) 1996-1997 Peter Eriksson <pen@lysator.liu.se>

   Version 1.3

   This program converts a directory tree from K/A-Share
   filename format to use ISO 8859-1 characters.
   
   Syntax:  kas_convfs [-n] [-u] <dir> [... <dir-N>]

       -n ..... Don't actually rename anything. Just print out
       		what would have been changed if this had been
		a real run.

       -u ..... Undo a tree translation.

       -f ..... Force translation (even though the .HSResource/*
		resource fork of a file doesn't exist)						
   This program is free software; you can redistribute it
   and/or modify it any way you like - as long as you write down all
   the modifications to it in this file, along with your name and
   email address. 
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include <stdio.h>
#include <dirent.h>
#include <ftw.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>


int dryrun;
int undo;
int safer;


/* Conversion table from `mac' charset to `latin1' charset.
   Generated mechanically by GNU recode 3.4.

   The recoding should be reversible.
*/

static unsigned char const mac_to_latin1[256] =
  {
      0,   1,   2,   3,   4,   5,   6,   7,	/*   0 -   7 */
      8,   9,  10,  11,  12,  13,  14,  15,	/*   8 -  15 */
     16,  17,  18,  19,  20,  21,  22,  23,	/*  16 -  23 */
     24,  25,  26,  27,  28,  29,  30,  31,	/*  24 -  31 */
     32,  33,  34,  35,  36,  37,  38,  39,	/*  32 -  39 */
     40,  41,  42,  43,  44,  45,  46,  47,	/*  40 -  47 */
     48,  49,  50,  51,  52,  53,  54,  55,	/*  48 -  55 */
     56,  57,  58,  59,  60,  61,  62,  63,	/*  56 -  63 */
     64,  65,  66,  67,  68,  69,  70,  71,	/*  64 -  71 */
     72,  73,  74,  75,  76,  77,  78,  79,	/*  72 -  79 */
     80,  81,  82,  83,  84,  85,  86,  87,	/*  80 -  87 */
     88,  89,  90,  91,  92,  93,  94,  95,	/*  88 -  95 */
     96,  97,  98,  99, 100, 101, 102, 103,	/*  96 - 103 */
    104, 105, 106, 107, 108, 109, 110, 111,	/* 104 - 111 */
    112, 113, 114, 115, 116, 117, 118, 119,	/* 112 - 119 */
    120, 121, 122, 123, 124, 125, 126, 127,	/* 120 - 127 */
    196, 197, 199, 201, 209, 214, 220, 225,	/* 128 - 135 */
    224, 226, 228, 227, 229, 231, 233, 232,	/* 136 - 143 */
    234, 235, 237, 236, 238, 239, 241, 243,	/* 144 - 151 */
    242, 244, 246, 245, 250, 249, 251, 252,	/* 152 - 159 */
    190, 176, 162, 163, 167, 130, 182, 223,	/* 160 - 167 */
    174, 169, 142, 180, 168, 173, 198, 216,	/* 168 - 175 */
    141, 177, 178, 179, 165, 181, 166, 135,	/* 176 - 183 */
    159, 185, 188, 170, 186, 189, 230, 248,	/* 184 - 191 */
    191, 161, 172, 146, 128, 129, 140, 171,	/* 192 - 199 */
    187, 131, 160, 192, 195, 213, 145, 147,	/* 200 - 207 */
    208, 132, 150, 148, 149, 144, 247, 215,	/* 208 - 215 */
    255, 153, 152, 164, 134, 221, 222, 151,	/* 216 - 223 */
    136, 183, 137, 139, 138, 194, 202, 193,	/* 224 - 231 */
    203, 200, 205, 206, 207, 204, 211, 212,	/* 232 - 239 */
    240, 210, 218, 219, 217, 155, 154, 133,	/* 240 - 247 */
    175, 157, 156, 158, 184, 253, 254, 143,	/* 248 - 255 */
  };

/* Conversion table from `latin1' charset to `mac' charset.
   Generated mechanically by GNU recode 3.4.

   The recoding should be reversible.
*/

static unsigned char const latin1_to_mac[256] =
  {
      0,   1,   2,   3,   4,   5,   6,   7,	/*   0 -   7 */
      8,   9,  10,  11,  12,  13,  14,  15,	/*   8 -  15 */
     16,  17,  18,  19,  20,  21,  22,  23,	/*  16 -  23 */
     24,  25,  26,  27,  28,  29,  30,  31,	/*  24 -  31 */
     32,  33,  34,  35,  36,  37,  38,  39,	/*  32 -  39 */
     40,  41,  42,  43,  44,  45,  46,  47,	/*  40 -  47 */
     48,  49,  50,  51,  52,  53,  54,  55,	/*  48 -  55 */
     56,  57,  58,  59,  60,  61,  62,  63,	/*  56 -  63 */
     64,  65,  66,  67,  68,  69,  70,  71,	/*  64 -  71 */
     72,  73,  74,  75,  76,  77,  78,  79,	/*  72 -  79 */
     80,  81,  82,  83,  84,  85,  86,  87,	/*  80 -  87 */
     88,  89,  90,  91,  92,  93,  94,  95,	/*  88 -  95 */
     96,  97,  98,  99, 100, 101, 102, 103,	/*  96 - 103 */
    104, 105, 106, 107, 108, 109, 110, 111,	/* 104 - 111 */
    112, 113, 114, 115, 116, 117, 118, 119,	/* 112 - 119 */
    120, 121, 122, 123, 124, 125, 126, 127,	/* 120 - 127 */
    196, 197, 165, 201, 209, 247, 220, 183,	/* 128 - 135 */
    224, 226, 228, 227, 198, 176, 170, 255,	/* 136 - 143 */
    213, 206, 195, 207, 211, 212, 210, 223,	/* 144 - 151 */
    218, 217, 246, 245, 250, 249, 251, 184,	/* 152 - 159 */
    202, 193, 162, 163, 219, 180, 182, 164,	/* 160 - 167 */
    172, 169, 187, 199, 194, 173, 168, 248,	/* 168 - 175 */
    161, 177, 178, 179, 171, 181, 166, 225,	/* 176 - 183 */
    252, 185, 188, 200, 186, 189, 160, 192,	/* 184 - 191 */
    203, 231, 229, 204, 128, 129, 174, 130,	/* 192 - 199 */
    233, 131, 230, 232, 237, 234, 235, 236,	/* 200 - 207 */
    208, 132, 241, 238, 239, 205, 133, 215,	/* 208 - 215 */
    175, 244, 242, 243, 134, 221, 222, 167,	/* 216 - 223 */
    136, 135, 137, 139, 138, 140, 190, 141,	/* 224 - 231 */
    143, 142, 144, 145, 147, 146, 148, 149,	/* 232 - 239 */
    240, 150, 152, 151, 153, 155, 154, 214,	/* 240 - 247 */
    191, 157, 156, 158, 159, 253, 254, 216,	/* 248 - 255 */
  };



static int hex2num(int c)
{
    if (isdigit(c))
	return c - '0';
    else if (c >= 'A' && c <= 'F')
	return c - 'A' + 10;
    else if (c >= 'a' && c <= 'f')
	return c - 'a' + 10;
    else
	return -1;
}


/* Convert ":XX" -> ISO characters.
**
*/
static void kas2iso(const char *in, char *out, int size)
{
    char *end = out + size - 4;
    
    while (*in && out < end)
    {
	if (in[0] == ':' && in[1] == ':')
	{
	    *out++ = *in++;
	    *out++ = *in++;
	}
	else if (in[0] == ':' && in[1] && in[2] &&
	    (isxdigit(in[1]) && !islower(in[1]) &&
	     isxdigit(in[2]) && !islower(in[2])))
	{
	    int code;

	    
	    code = mac_to_latin1[hex2num(in[1])*16+hex2num(in[2])];

	    if (code == '/')
	    {
		*out++ = ':';
		*out++ = ':';
	    }
	    else
		*out++ = code;

	    in += 3;
	}
	else
	    *out++ = *in++;
    }

    *out = '\0';
}


static void iso2kas(const unsigned char *in, char *out, int size)
{
    char *end = out + size - 4;
    
    while (*in && out < end)
    {
	if (in[0] == ':' && in[1] == ':')
	{
	    sprintf(out, ":%02X", '/');
	    out += 3;
	    in += 2;
	}
	else if (in[0] == ':' && isxdigit(in[1]) && isxdigit(in[2]))
	{
	    *out++ = *in++;
	    *out++ = *in++;
	    *out++ = *in++;
	}	
	else if (*in < ' ' || *in >= 128 || *in == ':')
	{
	    int code = latin1_to_mac[*in];
	    sprintf(out, ":%02X", code);
	    out += 3;
	    in++;
	}
	else
	    *out++ = *in++;
    }

    *out = '\0';
}


int walker(const char *path,
	   const struct stat *sp,
	   int flags,
	   struct FTW *fp)
{
    char buf[2048];
    char buf2[2048], buf3[2048];
    int code = 0;
    

    if (undo)
    {
	strcpy(buf, path);
	iso2kas(path+fp->base, buf+fp->base, sizeof(buf)-fp->base);
	if (strcmp(path+fp->base, buf+fp->base) == 0)
	    return 0;

	if (!dryrun && (code = rename(path, buf)) < 0)
	{
	    perror(path);
	    return 0;
	}
    }
    else
    {
	strcpy(buf, path);
	kas2iso(path+fp->base, buf+fp->base, sizeof(buf)-fp->base);
	
	if (strcmp(path+fp->base, buf+fp->base) == 0)
	    return 0;

	if (safer)
	{
	    strncpy(buf2, path, fp->base);
	    strcpy(buf2+fp->base, ".HSResource/");

	    strcpy(buf3, buf2);
	    
	    strcat(buf2+fp->base, path+fp->base);

	    strcat(buf3, buf+fp->base);

	    if (dryrun)
		code = access(buf2, R_OK);
	    else
		code = rename(buf2, buf3);

	    if (code == 0)
		printf("%s -> %s\n", buf2, buf3);
	}
	else
	    code = 0;

	if (!dryrun && code == 0 && (code = rename(path, buf)) < 0)
	{
	    perror(path);
	    return 0;
	}

    }

    if (code == 0)
	printf("%s -> %s\n", path, buf);
    
    return 0;
}


main(int argc, char *argv[])
{
    struct rlimit rlb;
    char buf[2048];
    DIR *dp;
    struct dirent *dep;
    int i;

    getrlimit(RLIMIT_NOFILE, &rlb);
    rlb.rlim_cur = rlb.rlim_max;
    setrlimit(RLIMIT_NOFILE, &rlb);

    dryrun = 0;
    undo = 0;
    safer = 1;
    
    for (i = 1; i < argc && argv[i][0] == '-'; i++)
	switch (argv[i][1])
	{
	case 'n':
	    dryrun = 1;
	    break;

	case 'f':
	    safer = 0;
	    break;
	    
	case 'u':
	    undo = 1;
	    break;

	case 'h':
	    printf("Syntax: %s [-u] [-n] [-f] <dir> [..<dir-N>]\n", argv[0]);
	    exit(0);

	default:
	    fprintf(stderr, "%s: unknown option: %s\n", argv[0], argv[i]);
	    exit(1);
	}

    while (i < argc)
	nftw(argv[i++], walker, 1024, FTW_DEPTH|FTW_MOUNT|FTW_PHYS);

    exit(0);
}
